(* Das Beispiel wurde dem Tutorial (C++) von www.snorre-dev.com entnommen und
   nach Delphi bersetzt
   J. Rathlev, Jan. 2002, info@rathlev-home.de *)

unit Lesson1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Direct3D8, d3dx8;

type
  TSample3DForm = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private-Deklarationen }
    // Das main Direct3D Interface, es wird zum initialisieren und schlieen von D3D bentigt
    lpd3d            : IDIRECT3D8;

    // Das D3DDevice wird zum Rendern benutzt und spiegelt den Bildschirm mit allen
    // Funktionen wieder. Wenn wir das D3DRender Interface erstellen, verndern oder 
    // das Bild rendern, so machen wir das ber dieses Interface
    lpd3ddevice      : IDirect3DDevice8;

    red,green,blue   : byte;

    procedure FatalError(hr : HResult; FehlerMsg : string);
    function D3DFind16BitMode : TD3DFORMAT;
    procedure D3DInit;
    procedure D3DShutdown;
    procedure D3DInitScene;
    procedure D3DKillScene;
    procedure D3DRender;
  public
    { Public-Deklarationen }
  end;

var
  Sample3DForm: TSample3DForm;

implementation

{$R *.DFM}

// Initialisieren aller Methoden und Variablen
procedure TSample3DForm.FormCreate(Sender: TObject);
begin
  lpd3d:=nil;
  lpd3ddevice:=nil;
  red:=0; green:=0; blue:=0;
  end;

procedure TSample3DForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  // Lsche die D3D Scene bevor wir D3D beenden
  D3DKillScene;
  // Lsche D3D
  D3DShutdown;
  end;

procedure TSample3DForm.FormShow(Sender: TObject);
begin
  D3DInit;     // Initialisieren von D3D
  D3DInitScene;
  D3DRender;   // Zeichne unsere Grafiken
  end;

procedure TSample3DForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_ESCAPE then close;
  end;

// Fataler Fehler. Meldung und Programmende
procedure TSample3DForm.FatalError(hr : HResult; FehlerMsg : string);
var
  s : string;
begin
  if hr<>0 then s:=D3DXErrorString(hr)+#13+FehlerMsg else s:=FehlerMsg;
  D3DKillScene;
  D3DShutdown;
  MessageDlg(s,mtError,[mbOK],0);
  close;
  end;

function TSample3DForm.D3DFind16BitMode : TD3DFORMAT;
var
  hr    : HRESULT;
begin
  hr:=lpd3d.CheckDeviceType (D3DADAPTER_DEFAULT,
                             D3DDEVTYPE_HAL,
                             D3DFMT_R5G6B5,      // Format Primary
                             D3DFMT_R5G6B5,      // Format Buffer
                             FALSE);
  if (SUCCEEDED(hr)) then begin
    Result:=D3DFMT_R5G6B5; exit;
    end;

  hr:=lpd3d.CheckDeviceType (D3DADAPTER_DEFAULT,
                             D3DDEVTYPE_HAL,
                             D3DFMT_X1R5G5B5,
                             D3DFMT_X1R5G5B5,
                             FALSE);
  if (SUCCEEDED(hr)) then begin
    Result:=D3DFMT_X1R5G5B5; exit;
    end;
  FatalError(0,'Der erforderliche D3D Modus wird von Ihrer Karte nicht untersttzt');
  Result:=0;
  end;

// Mit dieser Funktion initialisieren wir D3D
procedure TSample3DForm.D3DInit;
var
  hr    : HRESULT;
  d3dpp : TD3DPRESENTPARAMETERS;
begin
  //Erstelle Direct3D! Mu immer als erstes erstellt werden
  //Immer D3D_SDK_VERSION als Version setzen
  lpd3d:=Direct3DCreate8(D3D_SDK_VERSION);
  if(lpd3d=nil) then FatalError(0,'Fehler beim Erstellen von Direct3D!');

  // Setze D3DPRESENT_PARAMETERS auf 0, sonst knnten wir Probleme mit lteren
  // Eintragungen bzw. unkontrollierbaren Ergebnissen bekommen!
  // Sollten wir bei allen DirectX Strukturen machen
  ZeroMemory(@d3dpp,sizeof(d3dpp));

  with d3dpp do begin
    // Hiermit werden alte Frames gelscht, denn wir brauchen sie nicht
    SwapEffect:=D3DSWAPEFFECT_DISCARD;
    hDeviceWindow:=Handle;      // Dies ist unser HWND von TForm
    BackBufferCount:=1;         // 1 Backbuffer

    Windowed          := FALSE;
    BackBufferWidth   := 640;
    BackBufferHeight  := 480;
    BackBufferFormat  := D3DFind16BitMode;
    end;

  //Nachdem wir die D3DPRESENT_PARAMETERS Struktur ausgefllt haben, sind wir
  // endlich so weit unser D3D Device zu erstellen
  hr:=lpd3d.CreateDevice(D3DADAPTER_DEFAULT,
                         D3DDEVTYPE_HAL,
                         Handle,
                         D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                         d3dpp,
                         lpd3ddevice);
  if FAILED(hr) then FatalError(hr,'Fehler beim Erzeugen des 3D-Device');
  end;

// *** D3DShutdown hier werden die Resourcen von D3D wieder freigegeben
procedure TSample3DForm.D3DShutdown;
begin
  if assigned(lpd3ddevice) then lpd3ddevice:=nil;
  if assigned(lpd3d) then lpd3d:=nil;
  end;

procedure TSample3DForm.D3DInitScene;
begin
  end;

procedure TSample3DForm.D3DKillScene;
begin
  end;

procedure TSample3DForm.D3DRender;
begin
  inc(blue);
  if assigned(lpd3ddevice) then with lpd3ddevice do begin
    Clear(0,           // Wieviel Rechtecke lschen? 0 Lscht alle
          nil,         //Pointer zu den Rechtecken. NULL = Ganzer Bildschirm
          D3DCLEAR_TARGET,
          D3DCOLOR_XRGB(red,green,blue), //Hintergrundfarbe
          1,           // Lsche ZBuffer ( Wir haben momentan noch keinen )
          0 );

    if SUCCEEDED(BeginScene) then begin
      // Hier wird spter alles stehen, was gerendert werden soll

      EndScene;
      end;

   // Zeige Resultate auf dem Bildschirm
    Present(nil,nil,0,nil);
    end;
  end;

end.
