(* Das Beispiel wurde dem Tutorial (C++) von www.snorre-dev.com entnommen und
   nach Delphi bersetzt
   J. Rathlev, Jan. 2002, info@rathlev-home.de *)

unit Lesson2;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Direct3D8, d3dx8, ExtCtrls;

type
// Unsere Struktur, in der wir die Dreiecke speichern
// siehe dazu "Transformed and Lit Vertices" in MS-DirectX 8.0 SDK
  TMyVertex = record
    x,y,z,rhw : single;   // Position des Vertex
    color     : dword;    // Farbe des Vertex
    end;

  TTriangleVertex = array [0..2] of TMyVertex;
  TQuadratVertex  = array [0..3] of TMyVertex;

const
// Beschreibung des Vertextyps: Mit D3DFVF_DIFFUSE sagen wir DX, das unsere
// Struktur eine Farbe hat. D3DFVF_XYZRHW bedeutet, dass es sich um ein transformiertes
// Vertex handelt
  D3D8T_CUSTOMVERTEX =D3DFVF_XYZRHW or D3DFVF_DIFFUSE;

  // Dreieck
  // x -> Rechts, y -> Oben, z -> Tiefe
  TriangleVertex : TTriangleVertex = (
    (x : 175.0; y :  50.0; z : 0.5; rhw : 1.0; color : $000000FF), // x, y, z, rhw, Farbe
    (x : 300.0; y : 300.0; z : 0.5; rhw : 1.0; color : $FFFF0000),
    (x :  50.0; y : 300.0; z : 0.5; rhw : 1.0; color : $00FFFFFF));

  // Quadrat
  // Leider werden keine Vierecke von D3D untersttzt so wie in OpenGL, also mssen wir
  // doch alles aus Vertices erstellen
  QuadratVertex : TQuadratVertex = (
    (x : 350.0; y : 300.0; z : 0.5; rhw : 1.0; color : $000000FF), // x, y, z, rhw, Farbe
    (x : 350.0; y :  50.0; z : 0.5; rhw : 1.0; color : $00FF0000),
    (x : 590.0; y : 300.0; z : 0.5; rhw : 1.0; color : $0000FFFF), // x, y, z, rhw, Farbe
    (x : 590.0; y :  50.0; z : 0.5; rhw : 1.0; color : $0000FF00));

type
  TSample3DForm = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private-Deklarationen }
    // Das main Direct3D Interface, es wird zum initialisieren und schlieen von D3D bentigt
    lpd3d            : IDIRECT3D8;

    // Das D3DDevice wird zum Rendern benutzt und spiegelt den Bildschirm mit allen
    // Funktionen wieder. Wenn wir das D3DRender Interface erstellen, verndern oder 
    // das Bild rendern, so machen wir das ber dieses Interface
    lpd3ddevice      : IDirect3DDevice8;

    // Die Buffer, die unsere Vertizes enthalten
    dxtriangle,dxsquare : IDirect3DVertexBuffer8;

    red,green,blue   : byte;

    procedure FatalError(hr : HResult; FehlerMsg : string);
    function D3DFind16BitMode : TD3DFORMAT;
    procedure D3DInit;
    procedure D3DShutdown;
    procedure D3DInitScene;
    procedure D3DKillScene;
    procedure D3DRender;
    procedure MyIdleHandler (Sender: TObject; var Done: Boolean);
  public
    { Public-Deklarationen }
  end;

var
  Sample3DForm: TSample3DForm;

implementation

{$R *.DFM}

// Initialisieren aller Variablen
procedure TSample3DForm.FormCreate(Sender: TObject);
begin
  lpd3d:=nil;
  lpd3ddevice:=nil;
  dxtriangle:=nil; dxsquare:=nil;
  red:=0; green:=0; blue:=0;
  Application.OnIdle:= MyIdleHandler;
  end;

// Fataler Fehler. Meldung und Programmende
procedure TSample3DForm.FatalError(hr : HResult; FehlerMsg : string);
var
  s : string;
begin
  if hr<>0 then s:=D3DXErrorString(hr)+#13+FehlerMsg else s:=FehlerMsg;
  D3DKillScene;
  D3DShutdown;
  MessageDlg(s,mtError,[mbOK],0);
  close;
  end;

function TSample3DForm.D3DFind16BitMode : TD3DFORMAT;
var
  hr    : HRESULT;
begin
  hr:=lpd3d.CheckDeviceType (D3DADAPTER_DEFAULT,
                             D3DDEVTYPE_HAL,
                             D3DFMT_R5G6B5,      // Format Primary
                             D3DFMT_R5G6B5,      // Format Buffer
                             FALSE);
  if (SUCCEEDED(hr)) then begin
    Result:=D3DFMT_R5G6B5; exit;
    end;

  hr:=lpd3d.CheckDeviceType (D3DADAPTER_DEFAULT,
                             D3DDEVTYPE_HAL,
                             D3DFMT_X1R5G5B5,
                             D3DFMT_X1R5G5B5,
                             FALSE);
  if (SUCCEEDED(hr)) then begin
    Result:=D3DFMT_X1R5G5B5; exit;
    end;
  FatalError(0,'Der erforderliche D3D Modus wird von Ihrer Karte nicht untersttzt');
  Result:=0;
  end;

// Mit dieser Funktion initialisieren wir D3D
procedure TSample3DForm.D3DInit;
var
  hr    : HRESULT;
  d3dpp : TD3DPRESENTPARAMETERS;
begin
  //Erstelle Direct3D! Mu immer als erstes erstellt werden
  //Immer D3D_SDK_VERSION als Version setzen
  lpd3d:=Direct3DCreate8(D3D_SDK_VERSION);
  if(lpd3d=nil) then FatalError(0,'Fehler beim Erstellen von Direct3D!');

  // Setze D3DPRESENT_PARAMETERS auf 0, sonst knnten wir Probleme mit lteren
  // Eintragungen bzw. unkontrollierbaren Ergebnissen bekommen!
  // Sollten wir bei allen DirectX Strukturen machen
  ZeroMemory(@d3dpp,sizeof(d3dpp));

  with d3dpp do begin
    // Hiermit werden alte Frames gelscht, denn wir brauchen sie nicht
    SwapEffect:=D3DSWAPEFFECT_DISCARD;
    hDeviceWindow:=Handle;      // Dies ist unser HWND von TForm
    BackBufferCount:=1;         // 1 Backbuffer

    Windowed          := FALSE;
    BackBufferWidth   := 640;
    BackBufferHeight  := 480;
    BackBufferFormat  := D3DFind16BitMode;
    end;

  //Nachdem wir die D3DPRESENT_PARAMETERS Struktur ausgefllt haben, sind wir
  // endlich so weit unser D3D Device zu erstellen
  hr:=lpd3d.CreateDevice(D3DADAPTER_DEFAULT,
                         D3DDEVTYPE_HAL,
                         Handle,
                         D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                         d3dpp,
                         lpd3ddevice);
  if FAILED(hr) then FatalError(hr,'Fehler beim Erzeugen des 3D-Device');
  end;

// *** D3DShutdown hier werden die Resourcen von D3D wieder freigegeben
procedure TSample3DForm.D3DShutdown;
begin
  if assigned(lpd3ddevice) then lpd3ddevice:=nil;
  if assigned(lpd3d) then lpd3d:=nil;
  end;

procedure TSample3DForm.D3DInitScene;
var
  hr         : HRESULT;
  vbVertices : pByte;
begin
  if assigned(lpd3ddevice) then with lpd3ddevice do begin
    // Hier wird der Vertex Buffer fr das Dreieck erstellt
    hr:=CreateVertexBuffer (sizeof(TTriangleVertex),
                            D3DUSAGE_WRITEONLY, // Nur Schreibzugriffe
                            D3D8T_CUSTOMVERTEX, // Unser Vertex
                            D3DPOOL_MANAGED,
                            dxtriangle);       // Pointer zu unserem Dreieck

    if FAILED(hr) then FatalError(0,'Fehler beim Erstellen des Vertex Buffers fr unser Dreieck');

    // Hier wird der Vertex Buffer fr das Quadrat erstellt
    hr:=CreateVertexBuffer (sizeof(TQuadratVertex),
                            D3DUSAGE_WRITEONLY,
                            D3D8T_CUSTOMVERTEX,
                            D3DPOOL_MANAGED,
                            dxsquare);

    if FAILED(hr) then FatalError(0,'Fehler beim Erstellen des Vertex Buffers fr unser Viereck');

    // Nun kopieren wir unsere Vertizes in den Buffer
    // Wir mssen es zuvor mit Lock festhalten, um es bearbeiten zu knnen
    with dxtriangle do begin
      hr:=Lock(0, // Offset, an dem wir beginnen
               0, // Gre des locks ( 0 fr alles )
               vbVertices, // Wenn erfolgreich dann hier ablegen
               0); // sonstige Flags
      if FAILED(hr) then FatalError(0,'Fehler beim Locken des Buffers fr die Dreiecke');
      // Hier wird der Vertexbuffer kopiert.
      Move(TriangleVertex,vbVertices^,SizeOf(TTriangleVertex));
      // Und wieder loslassen
      Unlock;
      end;

    // Dasselbe Spiel mit dem Viereck
    with dxsquare do begin
      hr:=Lock(0,0,vbVertices,0);
      if FAILED(hr) then FatalError(0,'Fehler beim Locken des Buffers fr die Quadrate');
      Move(QuadratVertex,vbVertices^,SizeOf(TQuadratVertex));
      Unlock;
      end;
    end;
  end;    

procedure TSample3DForm.D3DKillScene;
begin
  dxtriangle:=nil;
  dxsquare:=nil;
  end;

procedure TSample3DForm.D3DRender;
begin
  // Verndern der Hintergrundfarbe
  if blue<255 then inc(blue)
  else if green<255 then inc(green)
  else if red<255 then inc(red) else begin
    red:=0; green:=0; blue:=0;
    end;

  if assigned(lpd3ddevice) then with lpd3ddevice do begin
    Clear(0,           // Wieviel Rechtecke lschen? 0 Lscht alle
          nil,         // Pointer zu den Rechtecken. nil = Ganzer Bildschirm
          D3DCLEAR_TARGET,
          D3DCOLOR_XRGB(red,green,blue), //Hintergrundfarbe
          1,           // Lsche ZBuffer ( Wir haben momentan noch keinen )
          0 );

    if SUCCEEDED(BeginScene) then begin
      // Vertex Shader sind wirklich komplex, aber es lassen sich damit gute Effekte
      // erzielen. Genauere Beschreibungen in der SDK, denn alles hier niederschreiben
      // sprengt den Rahmen eines Tutorials
      SetVertexShader(D3D8T_CUSTOMVERTEX);

      // Die D3D Renderfunktionen lesen aus Streams. Hier sagen wir DX welchen Stream
      // es verwenden soll
      SetStreamSource(0,dxtriangle,sizeof(TMyVertex));

      // Hier zeichnen wir nun endlich unser Dreieck. Wir bergeben DirectX, was wir
      // zeichen wollen, 0 fr den Index des 1. Vertex, 1 fr die Anzahl der Vertizes
      DrawPrimitive(D3DPT_TRIANGLELIST,0,1);

      // Jetzt setzen wir den Stream auf unser Quadrat
      SetStreamSource(0,dxsquare,sizeof(TMyVertex));

      // Hier werden die 2 Dreiecke gezeichnet
      DrawPrimitive(D3DPT_TRIANGLESTRIP,0,2);

      EndScene;
      end;

   // Zeige Resultate auf dem Bildschirm
    Present(nil,nil,0,nil);
    end;
  end;

procedure TSample3DForm.FormShow(Sender: TObject);
begin
  D3DInit;     // Initialisieren von D3D
  D3DInitScene;
  D3DRender;   // Zeichne unsere Grafiken
  end;

procedure TSample3DForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  // Lsche die D3D Scene bevor wir D3D beenden
  D3DKillScene;
  // Lsche D3D
  D3DShutdown;
  end;

procedure TSample3DForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_ESCAPE then close;
  if Key=VK_SPACE then D3DRender;
  end;

procedure TSample3DForm.MyIdleHandler (Sender: TObject; var Done: Boolean);
begin
  D3DRender;
  Done:=false;
  end;

end.
