(* Lektion 6: Benutzung des Z-Buffers, Einfgen eines Hintergrunds
              Vollbild oder Fenstermodus
   (siehe auch C++ Beispiele von www.snorre-dev.com)
   Die Texturen fr den Hintergrund wurden den Beispielen von
   Microsoft zu DirectX entnommen (Bestandteil des SDK).
   J. Rathlev, Feb. 2002, info@rathlev-home.de *)

unit Lesson6;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Direct3D8, d3dx8, ExtCtrls;

const
  Pi180 = Pi/180.0;
  MPath = 'Media\';       // Pfad mit Texturen
  Fovy = Pi/4;            // ffnungswinkel des Sichtkegels (Field of view)
  ViewDist = 15;          // Abstand des Betrachters vom Nullpunkt
  ViewHeight = 5;         // Hhe des Betrachters
  BoxDepth = 0.5;         // Eintauchtiefe der Kiste
  Delta = 0.3;            // Winkelinkrement fr Animation

  CubeScale = 1.5;        // Skalierungsfaktor fr Kiste
  SkyScale = 100;         // Skalierungsfaktor fr Hintergrundbox
  CubeCount = 36;

type
  // Unsere Struktur, in der wir die Dreiecke speichern
  TMyVertex = record
    x,y,z     : single;   // Position des Vertex
    color     : dword;    // Farbe des Vertex
    tu,tv     : single;   // Die Koordinaten der Textur
    end;

  // Vertexliste fr Wrfel
  TCube = array [0..CubeCount-1] of TMyVertex;

const
  // Beschreibung des Vertextyps (siehe Lektion 5)
  D3D8T_CUSTOMVERTEX =D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEX1;

  // Standardwrfel (Seitenlnge 2)
  NormCube : TCube = (
    (x :-1.0; y :-1.0; z :-1.0; color : $000000FF; tu : 0; tv : 1),  // Vorn
    (x :-1.0; y : 1.0; z :-1.0; color : $000000FF; tu : 0; tv : 0),
    (x : 1.0; y : 1.0; z :-1.0; color : $000000FF; tu : 1; tv : 0),
    (x : 1.0; y : 1.0; z :-1.0; color : $000000FF; tu : 1; tv : 0),
    (x : 1.0; y :-1.0; z :-1.0; color : $000000FF; tu : 1; tv : 1),
    (x :-1.0; y :-1.0; z :-1.0; color : $000000FF; tu : 0; tv : 1),

    (x :-1.0; y :-1.0; z : 1.0; color : $00FF00FF; tu : 0; tv : 1),  // Links
    (x :-1.0; y : 1.0; z : 1.0; color : $00FF00FF; tu : 0; tv : 0),
    (x :-1.0; y : 1.0; z :-1.0; color : $00FF00FF; tu : 1; tv : 0),
    (x :-1.0; y : 1.0; z :-1.0; color : $00FF00FF; tu : 1; tv : 0),
    (x :-1.0; y :-1.0; z :-1.0; color : $00FF00FF; tu : 1; tv : 1),
    (x :-1.0; y :-1.0; z : 1.0; color : $00FF00FF; tu : 0; tv : 1),

    (x : 1.0; y :-1.0; z : 1.0; color : $0000FF00; tu : 0; tv : 1),  // Hinten
    (x : 1.0; y : 1.0; z : 1.0; color : $0000FF00; tu : 0; tv : 0),
    (x :-1.0; y : 1.0; z : 1.0; color : $0000FF00; tu : 1; tv : 0),
    (x :-1.0; y : 1.0; z : 1.0; color : $0000FF00; tu : 1; tv : 0),
    (x :-1.0; y :-1.0; z : 1.0; color : $0000FF00; tu : 1; tv : 1),
    (x : 1.0; y :-1.0; z : 1.0; color : $0000FF00; tu : 0; tv : 1),

    (x : 1.0; y :-1.0; z :-1.0; color : $0000FFFF; tu : 0; tv : 1),  // Rechts
    (x : 1.0; y : 1.0; z :-1.0; color : $0000FFFF; tu : 0; tv : 0),
    (x : 1.0; y : 1.0; z : 1.0; color : $0000FFFF; tu : 1; tv : 0),
    (x : 1.0; y : 1.0; z : 1.0; color : $0000FFFF; tu : 1; tv : 0),
    (x : 1.0; y :-1.0; z : 1.0; color : $0000FFFF; tu : 1; tv : 1),
    (x : 1.0; y :-1.0; z :-1.0; color : $0000FFFF; tu : 0; tv : 1),

    (x :-1.0; y : 1.0; z :-1.0; color : $00FF0000; tu : 0; tv : 1),  // Oben
    (x :-1.0; y : 1.0; z : 1.0; color : $00FF0000; tu : 0; tv : 0),
    (x : 1.0; y : 1.0; z : 1.0; color : $00FF0000; tu : 1; tv : 0),
    (x : 1.0; y : 1.0; z : 1.0; color : $00FF0000; tu : 1; tv : 0),
    (x : 1.0; y : 1.0; z :-1.0; color : $00FF0000; tu : 1; tv : 1),
    (x :-1.0; y : 1.0; z :-1.0; color : $00FF0000; tu : 0; tv : 1),

    (x : 1.0; y :-1.0; z :-1.0; color : $00FFFF00; tu : 0; tv : 1),  // Unten
    (x : 1.0; y :-1.0; z : 1.0; color : $00FFFF00; tu : 0; tv : 0),
    (x :-1.0; y :-1.0; z : 1.0; color : $00FFFF00; tu : 1; tv : 0),
    (x :-1.0; y :-1.0; z : 1.0; color : $00FFFF00; tu : 1; tv : 0),
    (x :-1.0; y :-1.0; z :-1.0; color : $00FFFF00; tu : 1; tv : 1),
    (x : 1.0; y :-1.0; z :-1.0; color : $00FFFF00; tu : 0; tv : 1));

type
  TSample3DForm = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private-Deklarationen }
    // Direct3D Interfaces (siehe Lektion 5)
    lpd3d            : IDIRECT3D8;
    lpd3ddevice      : IDirect3DDevice8;
    D3dDevCaps       : TD3DCaps8;

    // Vertexbuffer fr Wrfel
    CubeVB : IDirect3DVertexBuffer8;

    // Unsere Texturen
    SkyTop,SkyBottom,
    SkyFront,SkyBack,
    SkyLeft,SkyRight,
    CubeTexture   : IDIRECT3DTEXTURE8;

    // Rotationswinkel fr den Beobachter in Grad
    RotY   : single  ;

    HwVertexProcess,
    FullScreen,
    Animate          : boolean;

    procedure FatalError(hr : HResult; FehlerMsg : string);
    function D3DFind16BitMode : TD3DFORMAT;
    procedure D3DInit;
    procedure D3DShutdown;
    procedure D3DInitScene;
    procedure D3DKillScene;
    procedure D3DRender;
    procedure MyIdleHandler (Sender: TObject; var Done: Boolean);
  public
    { Public-Deklarationen }
  end;

var
  Sample3DForm: TSample3DForm;

implementation

{$R *.DFM}

// Initialisieren aller Variablen
procedure TSample3DForm.FormCreate(Sender: TObject);
begin
  lpd3d:=nil;
  lpd3ddevice:=nil;
  CubeVB:=nil;
  CubeTexture:=nil;
  SkyTop:=nil; SkyBottom:=nil;
  SkyFront:=nil; SkyBack:=nil;
  SkyLeft:=nil; SkyRight:=nil;
  //Initialisiere die Rotation 
  RotY:=0;
  FullScreen:=false;     // setze auf"true" fr Vollbildmodus
  Animate:=false;
  Application.OnIdle:= MyIdleHandler;
  end;

// Fataler Fehler. Meldung und Programmende
procedure TSample3DForm.FatalError(hr : HResult; FehlerMsg : string);
var
  s : string;
begin
  if hr<>0 then s:=D3DXErrorString(hr)+#13+FehlerMsg else s:=FehlerMsg;
  D3DKillScene;
  D3DShutdown;
  MessageDlg(s,mtError,[mbOK],0);
  close;
  end;

function TSample3DForm.D3DFind16BitMode : TD3DFORMAT;
var
  hr    : HRESULT;
begin
  hr:=lpd3d.CheckDeviceType (D3DADAPTER_DEFAULT,
                             D3DDEVTYPE_HAL,
                             D3DFMT_R5G6B5,      // Format Primary
                             D3DFMT_R5G6B5,      // Format Buffer
                             FALSE);
  if (SUCCEEDED(hr)) then begin
    Result:=D3DFMT_R5G6B5; exit;
    end;

  hr:=lpd3d.CheckDeviceType (D3DADAPTER_DEFAULT,
                             D3DDEVTYPE_HAL,
                             D3DFMT_X1R5G5B5,
                             D3DFMT_X1R5G5B5,
                             FALSE);
  if (SUCCEEDED(hr)) then begin
    Result:=D3DFMT_X1R5G5B5; exit;
    end;
  FatalError(0,'Der erforderliche D3D Modus wird von Ihrer Karte nicht untersttzt');
  Result:=0;
  end;

// Mit dieser Funktion initialisieren wir D3D
procedure TSample3DForm.D3DInit;
var
  hr    : HRESULT;
  d3dpp : TD3DPRESENTPARAMETERS;
  d3ddm : TD3DDISPLAYMODE;
  vp    : integer;
begin
  //Erstelle Direct3D! Mu immer als erstes erstellt werden
  //Immer D3D_SDK_VERSION als Version setzen
  lpd3d:=Direct3DCreate8(D3D_SDK_VERSION);
  if(lpd3d=nil) then FatalError(0,'Fehler beim Erstellen von Direct3D!');

  // Setze zunchst alle D3DPRESENT_PARAMETERS auf 0
  ZeroMemory(@d3dpp,sizeof(d3dpp));
  with d3dpp do begin
    SwapEffect:=D3DSWAPEFFECT_DISCARD;
    hDeviceWindow:=Handle;      // Dies ist unser HWND von TForm

    // Wir brauche einen Z-Buffer also schalten wir ihn ein
    EnableAutoDepthStencil := TRUE;
    AutoDepthStencilFormat := D3DFMT_D16;

    // Initialisieren des Backbuffers
    Windowed:=not FullScreen;
    if FullScreen then begin
      // Vollbild
      BackBufferWidth   := 640;
      BackBufferHeight  := 480;
      BackBufferFormat  := D3DFind16BitMode;
      BackBufferCount:=1;         // 1 Backbuffer
      end
    else begin
      // Fenster
      hr:=lpd3d.GetAdapterDisplayMode(D3DADAPTER_DEFAULT,d3ddm);
      if failed(hr) then FatalError(hr,'Fehler beim Ermitteln des Dislaymodes');
      BackBufferFormat := d3ddm.Format;
      end;
    end;

  // Hardware T&L?
  hr:=lpd3d.GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,D3dDevCaps);
  if FAILED(hr) then FatalError(hr,'Fehler beim Abfragen der DevCaps');
  HwVertexProcess:=D3dDevCaps.DevCaps and D3DDEVCAPS_HWTRANSFORMANDLIGHT <> 0;

  if HwVertexProcess then vp:=D3DCREATE_HARDWARE_VERTEXPROCESSING
  else vp:=D3DCREATE_SOFTWARE_VERTEXPROCESSING;

  //Erstellen des D3D-Device
  hr:=lpd3d.CreateDevice(D3DADAPTER_DEFAULT,
                         D3DDEVTYPE_HAL,
                         Handle,
                         vp,
                         d3dpp,
                         lpd3ddevice);
  if FAILED(hr) then FatalError(hr,'Fehler beim Erzeugen des 3D-Device');
  end;

// *** D3DShutdown hier werden die Resourcen von D3D wieder freigegeben
procedure TSample3DForm.D3DShutdown;
begin
  if assigned(lpd3ddevice) then lpd3ddevice:=nil;
  if assigned(lpd3d) then lpd3d:=nil;
  end;

// Initialisieren der Szenenobjekte
procedure TSample3DForm.D3DInitScene;
var
  hr            : HRESULT;
  vbVertices    : pByte;
  ProjMatrix    : TD3DXMATRIX;

  // Textur laden
  function LoadTexture (Filename : string) : IDIRECT3DTEXTURE8;
  var
    hr         : HRESULT;
  begin
    if assigned(lpd3ddevice) then with lpd3ddevice do begin
      hr:=D3DXCreateTextureFromFile(lpd3ddevice,pchar(MPath+Filename),Result);
      if FAILED(hr) then FatalError(hr,'Fehler beim Laden der Textur: '+Filename);
      end;
    end;

begin
  if assigned(lpd3ddevice) then with lpd3ddevice do begin
    // Vertex Buffer fr Wrfel
    hr:=CreateVertexBuffer (sizeof(TCube),
                            D3DUSAGE_WRITEONLY,  // Nur Schreibzugriffe
                            D3D8T_CUSTOMVERTEX,  // Unser Vertex
                            D3DPOOL_MANAGED,
                            CubeVB);             // Pointer zu unserem Buffer

    if FAILED(hr) then FatalError(hr,'Fehler beim Erstellen des Vertex Buffers');

    // Nun kopieren wir unsere Vertizes in den Buffer
    // Wir mssen es zuvor mit Lock festhalten, um es bearbeiten zu knnen
    with CubeVB do begin
      hr:=Lock(0, // Offset, an dem wir beginnen
               0, // Gre des locks ( 0 fr alles )
               vbVertices, // Wenn erfolgreich dann hier ablegen
               0); // sonstige Flags
      if FAILED(hr) then FatalError(hr,'Fehler beim Locken des Vertex-Buffers');
      // Hier wird der Vertexbuffer kopiert.
      Move(NormCube,vbVertices^,SizeOf(TCube));
      // Und wieder loslassen
      Unlock;
      end;

    // Lighting abschalten, da unsere Vertices nocht keine Normalenvektoren besitzten
    SetRenderState(D3DRS_LIGHTING,0);
    // Z-Buffer beim Rendern einschalten
    SetRenderState(D3DRS_ZENABLE,D3DZB_TRUE);

    // Erstelle eine Projektionsmatrix
    D3DXMatrixPerspectiveFovLH(ProjMatrix,  // Resultierende Matrix
                               Fovy,        // ffnungswinkel des Sichtkegels
                               640/480,     // Seitenverhltnis der Ansicht
                               1.0,         // Mindeste Nhe
                               1000.0);     // Maximal sichtbare Entfernung
    SetTransform(D3DTS_PROJECTION,ProjMatrix );

    // Mit D3DTSS_COLORP wird festgelegt, wie die Farbe jedes einzelnen Pixels
    // verarbeitet wird. D3DTOP_SELECTARG1 verweist auf D3DTSS_COLORARG1
    SetTextureStageState(0,D3DTSS_COLOROP, D3DTOP_SELECTARG1);
    // Mit D3DTSS_COLORARG1 wird festgelegt, da die Farbe nur von der Textur
    // genommen wird und von nichts anderem.
    SetTextureStageState(0,D3DTSS_COLORARG1, D3DTA_TEXTURE);
    // Wir benutzen kein Alpha blending, also schalten wir es ab.
    SetTextureStageState(0,D3DTSS_ALPHAOP,   D3DTOP_DISABLE);
    // Texturfilter
    SetTextureStageState(0,D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
    SetTextureStageState(0,D3DTSS_MINFILTER, D3DTEXF_LINEAR);

    // Texturen laden
    CubeTexture:=LoadTexture ('ieap-kiste-g.bmp');
    SkyTop:=LoadTexture ('skybox_top.bmp');
    SkyBottom:=LoadTexture ('skybox_bottom.bmp');
    SkyFront:=LoadTexture ('skybox_front.bmp');
    SkyBack:=LoadTexture ('skybox_back.bmp');
    SkyLeft:=LoadTexture ('skybox_left.bmp');
    SkyRight:=LoadTexture ('skybox_right.bmp');
    end;
  Animate:=true;
  end;

procedure TSample3DForm.D3DKillScene;
begin
  CubeVB:=nil;
  CubeTexture:=nil;
  SkyTop:=nil; SkyBottom:=nil;
  SkyFront:=nil; SkyBack:=nil;
  SkyLeft:=nil; SkyRight:=nil;
  end;

// Rendern der Szene
procedure TSample3DForm.D3DRender;
var
  WorldMatrix,
  ViewMatrix,
  TempMatrix     : TD3DXMATRIX;
begin
  RotY:=RotY+Delta;   // Rotation des Beobachters

  if assigned(lpd3ddevice) then with lpd3ddevice do begin
    Clear(0,           // Wieviel Rechtecke lschen? 0 Lscht alle
          nil,         // Pointer zu den Rechtecken. nil = Ganzer Bildschirm
          D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
          D3DCOLOR_XRGB(0,0,0), //Hintergrundfarbe schwarz
          1,           // Lsche ZBuffer ( Wir haben momentan noch keinen )
          0 );

    if SUCCEEDED(BeginScene) then begin

      // Hier erstellen wir unsere SichtMatrix
      // Der Beobachter bewegt sich in einer Hhe von ViewHeight
      // kreisfrmig im Abstand "ViewDist" um die Kiste
      D3DXMatrixLookAtLH (ViewMatrix,D3DXVECTOR3(ViewDist*sin(Pi180*RotY),
                                                 ViewHeight,
                                                 ViewDist*cos(Pi180*RotY)),
                                     D3DXVECTOR3(0.0,0.0,0.0),
                                     D3DXVECTOR3(0.0,1.0,0.0));

      SetTransform(D3DTS_VIEW,ViewMatrix);

      // Vertex Typ einstellen
      SetVertexShader(D3D8T_CUSTOMVERTEX);

      // Stream auf Vertexbuffer fr Wrfel setzen
      SetStreamSource(0,CubeVB,sizeof(TMyVertex));

      // Berechne die Welt-Matrix fr den Hintergrund (Seiten und oben)
      D3DXMatrixScaling(WorldMatrix,SkyScale,SkyScale,SkyScale);
      SetTransform(D3DTS_WORLD,WorldMatrix);

      // Hintergrund (Seiten): Textur auswhlen und zeichnen
      SetRenderState(D3DRS_CULLMODE,D3DCULL_CW);
      SetTexture(0,SkyFront);         // Vorderseite des Hintergrunds
      DrawPrimitive(D3DPT_TRIANGLELIST,0,2);

      SetTexture(0,SkyLeft);         // linke Seite des Hintergrunds
      DrawPrimitive(D3DPT_TRIANGLELIST,6,2);

      SetTexture(0,SkyBack);         // Hinterseite des Hintergrunds
      DrawPrimitive(D3DPT_TRIANGLELIST,12,2);

      SetTexture(0,SkyRight);        // rechte Seite des Hintergrunds
      DrawPrimitive(D3DPT_TRIANGLELIST,18,2);

      // Hintergrund (oben): Textur auswhlen und zeichnen
      SetTexture(0,SkyTop);          // Oberseite des Hintergrunds
      DrawPrimitive(D3DPT_TRIANGLELIST,24,2);

      // Berechne die Welt-Matrix fr den Hintergrund (Boden)
      // setze dabei y auf 0 (Wasseroberflche)
      D3DXMatrixScaling(WorldMatrix,SkyScale,0,SkyScale);
      SetTransform(D3DTS_WORLD,WorldMatrix);
      // Hintergrund (unten): Textur auswhlen und zeichnen
      SetTexture(0,SkyBottom);
      DrawPrimitive(D3DPT_TRIANGLELIST,30,2);

      // Textur fr Kiste auswhlen
      SetTexture(0,CubeTexture);
      SetRenderState(D3DRS_CULLMODE,D3DCULL_CCW);
      // Setze die Welt-Matrix fr die Kiste etwas nach oben
      D3DXMatrixTranslation(WorldMatrix,0,1-BoxDepth,0);
      // Skaliere die Kiste
      D3DXMatrixScaling(TempMatrix,CubeScale,CubeScale,CubeScale);
      D3DXMatrixMultiply(WorldMatrix,WorldMatrix,TempMatrix);
      SetTransform(D3DTS_WORLD,WorldMatrix);

      // Zeichnen der Kiste
      DrawPrimitive(D3DPT_TRIANGLELIST,0,12);
      EndScene;
      end;

   // Zeige Resultate auf dem Bildschirm
    Present(nil,nil,0,nil);
    end;
  end;

procedure TSample3DForm.FormShow(Sender: TObject);
begin
  D3DInit;     // Initialisieren von D3D
  D3DInitScene;
  D3DRender;   // Darstellung der 3D-Szene
  end;

procedure TSample3DForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  // Lsche die 3D-Szene bevor wir D3D beenden
  D3DKillScene;
  // Lsche D3D
  D3DShutdown;
  end;

procedure TSample3DForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_ESCAPE then close;
  if Key=VK_SPACE then Animate:=not Animate;
  end;

procedure TSample3DForm.MyIdleHandler (Sender: TObject; var Done: Boolean);
begin
  if Animate then D3DRender;
  Done:=false;
  end;

end.
